/* 
 * Copyright (C) 2001-2007 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef LOGMANAGER_H
#define LOGMANAGER_H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "File.h"
#include "CriticalSection.h"
#include "Singleton.h"
#include "TimerManager.h"

class LogManagerListener {
public:
	virtual ~LogManagerListener() { }
	template<int I>	struct X { enum { TYPE = I };  };

	typedef X<0> Message;
	virtual void on(Message, time_t, const string&) throw() { }
};

class LogManager : public Singleton<LogManager>, public Speaker<LogManagerListener>
{
public:
	LogManager();
	virtual ~LogManager() throw() { }

	enum LogArea { CHAT, PM, DOWNLOAD, UPLOAD, SYSTEM, STATUS, LAST };
	enum {FILE, FORMAT};

	void log(LogArea area, StringMap& params) throw();
	deque<pair<time_t, string> > getLastLogs() { Lock l(cs); return lastLogs; }

	void message(const string& msg, bool logit = true);

	const string& getSetting(int area, int sel) {
		return SettingsManager::getInstance()->get(static_cast<SettingsManager::StrSetting>(logOptions[area][sel]), true);
	}

	void saveSetting(int area, int sel, const string& setting) {
		SettingsManager::getInstance()->set(static_cast<SettingsManager::StrSetting>(logOptions[area][sel]), setting);
	}

private:
	void log(const string& area, const string& msg) throw();

	friend class Singleton<LogManager>;
	CriticalSection cs;
	deque<pair<time_t, string> > lastLogs;
	
	int logOptions[LAST][2];	
};

#define LOG(area, msg) LogManager::getInstance()->log(area, msg)

#endif // LOGMANAGER_H

/* Some parts Copyright (C) 2002-2007 Petr Kozelka, PPK at czdc dot org */
