/*
 * Copyright (C) 2001-2007 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdinc.h"
#include "DCPlusPlus.h"

#include "NmdcHub.h"

#include "ResourceManager.h"
#include "ClientManager.h"
#include "SearchManager.h"
#include "ShareManager.h"
#include "CryptoManager.h"
#include "ConnectionManager.h"

#include "Socket.h"
#include "UserCommand.h"

NmdcHub::NmdcHub(const string& aHubURL) : Client(aHubURL, '|'), 
	lastbytesshared(0), validatenicksent(false), bFirstOpList(true), 
	PtokaX(false), ShittYnHub(false) { }

NmdcHub::~NmdcHub() throw() {
	clearUsers();
}

void NmdcHub::connect(const OnlineUser& aUser, const string&) {
	checkstate(); 
	dcdebug("NmdcHub::connect %s\n", aUser.getIdentity().getNick().c_str());
	if(ClientManager::getInstance()->isActive()) {
		connectToMe(aUser);
	} else {
		revConnectToMe(aUser);
	}
}

int64_t NmdcHub::getAvailable() const {
	Lock l(cs);
	int64_t x = 0;
	for(NickMap::const_iterator i = users.begin(); i != users.end(); ++i) {
		x+=i->second->getIdentity().getBytesShared();
	}
	return x;
}

OnlineUser& NmdcHub::getUser(const string& aNick) {
	OnlineUser* u = NULL;
	{
		Lock l(cs);

		NickIter i = users.find(aNick);
		if(i != users.end())
			return *i->second;
	}

	User::Ptr p;
	if(aNick == getCurrentNick()) {
		p = ClientManager::getInstance()->getMe();
	} else {
		p = ClientManager::getInstance()->getUser(aNick, getHubUrl());
	}
	
	{
		Lock l(cs);
		u = users.insert(make_pair(aNick, new OnlineUser(p, *this))).first->second;
		u->getIdentity().setNick(aNick);
		if(u->getUser() == getMyIdentity().getUser()) {
			setMyIdentity(u->getIdentity());
		}
	}

	ClientManager::getInstance()->putOnline(u);
	return *u;
}

void NmdcHub::supports(const StringList& feat) { 
	string x;
	for(StringList::const_iterator i = feat.begin(); i != feat.end(); ++i) {
		x+= ' ' + *i;
	}
	if(getStealth() == false) {
		 send("$Supports" + x + '|');
    } else {
		send("$Supports" + x + " |");
    }
}

OnlineUser* NmdcHub::findUser(const string& aNick) {
	Lock l(cs);
	NickIter i = users.find(aNick);
	return i == users.end() ? NULL : i->second;
}

void NmdcHub::putUser(const string& aNick) {
	OnlineUser* ou = NULL;
	{
		Lock l(cs);
		NickIter i = users.find(aNick);
		if(i == users.end())
			return;
		ou = i->second;
		users.erase(i);
	}
	ClientManager::getInstance()->putOffline(ou);
	delete ou;
}

void NmdcHub::clearUsers() {
	NickMap u2;

	{
		Lock l(cs);
		u2.swap(users);
	}

	for(NickIter i = u2.begin(); i != u2.end(); ++i) {
		ClientManager::getInstance()->putOffline(i->second);
		delete i->second;
	}
	u2.clear();
}

void NmdcHub::updateFromTag(Identity& id) {
    string tag = id.get("TA");
    if(tag == Util::emptyString)
        return;
    
    char *sTag = (char *)tag.c_str();
    sTag[tag.size()-1] = ',';
    char *sPart = sTag+1;
    bool bHaveSpace = false;
    for(size_t i = 1; i < tag.size(); i++) {
        switch(sTag[i]) {
            case ' ':
                if(bHaveSpace == false) {
                    if((sTag+i)-sPart > 0) {
                        string sCL = string(sPart, (sTag+i)-sPart);
                        id.set("CL", sCL);
                    }
                        
                    sPart = sTag+i+1;
                    bHaveSpace = true;
                }
                break;
            case ',':
                if(sPart[1] == ':') {
                    switch(sPart[0]) {
                		case 'V':
                            if((sTag+i)-(sPart+2)) {
                                string sVE = string(sPart+2, (sTag+i)-(sPart+2));
                                id.set("VE", sVE);
                            }
                                
                    		break;
                		case 'M':
                            if((sTag+i)-(sPart+2)) {
                                string sMD = string(sPart+2, (sTag+i)-(sPart+2));
                                id.set("MD", sMD);
                            }
                                
                			break;
                		case 'H':
                			if((sTag+i)-(sPart+2) > 4) {
                				int a,b,c;
                				if(sscanf_s(sPart+2, "%d/%d/%d", &a, &b, &c) == 3) {
                                    id.set("HN", Util::toString(a));
    			                    id.set("HR", Util::toString(b));
    			                    id.set("HO", Util::toString(c));
    			                    id.set("HT", Util::toString(a+b+c));
                				} else {
                                    if((sTag+i)-(sPart+2)) {
                                        string sHT = string(sPart+2, (sTag+i)-(sPart+2));
                                        id.set("HT", sHT);
                                    }
                                }
                			} else {
                                if((sTag+i)-(sPart+2)) {
                                    string sHT = string(sPart+2, (sTag+i)-(sPart+2));
                                    id.set("HT", sHT);
                                }
                			}
                    		break;
                		case 'S':
                            if((sTag+i)-(sPart+2)) {
                                string sSL = string(sPart+2, (sTag+i)-(sPart+2));
                                id.set("SL", sSL);
                            }
                                
                    		break;
                		case 'L':
                        case 'B':
                        case 'U':
                            if((sTag+i)-(sPart+2)) {
                                string sUL = string(sPart+2, (sTag+i)-(sPart+2));
                                id.set("UL", sUL);
                            }
                                
                			break;
                		default:
                       	    break; 
                    }
                }
                
                sPart = sTag+i+1;
                break;
            default:
                continue;
        }
    }
    sTag[tag.size()-1] = '>';
}

void NmdcHub::on(Line, const string& l) throw() {
    Client::on(Line(), l);
    
    COMMAND_DEBUG(l, DebugManager::HUB_IN, getIpPort());
  
    string line = toUtf8(l);
    char* aLine = (char *)line.c_str();
    int iaLineLen = line.size();
    
	if(aLine[0] != '$') {
    	// Check if we're being banned...
    	if(state != STATE_LOGGEDIN) {
    		if(strstr(aLine, "banned") != 0) {
    			setAutoReconnect(false);
    		}
    	}
    
    	bool bNormal = true;
    	OnlineUser* u = findUserInChat(aLine, bNormal);

        string sText = unescape(bNormal ? string(aLine, iaLineLen) : string(aLine+4, iaLineLen-4));
        fire(ClientListener::Message(), this, *u, sText);
        return;
    }

	switch(aLine[1]) {
		case 'Z':
			// $ZOn
			if(iaLineLen == 4 && ((uint16_t *)(aLine+2))[0] == ((uint16_t *)"On")[0]) {
				socket->setMode(BufferedSocket::MODE_ZPIPE);
				return;
			}
			
			break;
        case 'S':
        	// $Search
        	if(((uint32_t *)(aLine+2))[0] == ((uint32_t *)"earc")[0] && ((uint16_t *)(aLine+6))[0] == ((uint16_t *)"h ")[0]) {
        	    if(state != STATE_LOGGEDIN || iaLineLen < 19) return;
       	    
        	    char *sSearchParts[] = { NULL, NULL, NULL, NULL, NULL, NULL, NULL };
                unsigned short iSearchPartsLen[] = { 0, 0, 0, 0, 0, 0, 0 };
                
                unsigned char cPart = 0;
                
                sSearchParts[cPart] = aLine+8; // IP for active search, Hub for passive search...
                
        	    for(int i = 8; i < iaLineLen; i++) {
                    switch(aLine[i]) {
                        case ':':
                            if(cPart == 0) {
                                iSearchPartsLen[cPart] = (unsigned short)((aLine+i)-sSearchParts[cPart]);
                                cPart++;
                                sSearchParts[cPart] = aLine+i+1;
                            }
                            break;
                        case ' ':
                            if(cPart == 1) {
                                aLine[i] = '\0';
                                iSearchPartsLen[cPart] = (unsigned short)((aLine+i)-sSearchParts[cPart]);
                                cPart++;
                                sSearchParts[cPart] = aLine+i+1;
                            }
                            break;
                        case '?':
                            if(cPart > 1) {
                                aLine[i] = '\0';
                                iSearchPartsLen[cPart] = (unsigned short)((aLine+i)-sSearchParts[cPart]);
                                
                                // are we on last '?' ?
                                if(cPart == 5) {
                                    sSearchParts[6] = aLine+i+1;
                                    iSearchPartsLen[6] = (unsigned short)(iaLineLen-i-1);
                                    i = iaLineLen;
                                    break;
                                }
                                    
                                cPart++;
                                sSearchParts[cPart] = aLine+i+1;
                            }
                            break;
                    }
                }
       	    
       	        // check if we have all we need
        	    if(iSearchPartsLen[0] == 0 || iSearchPartsLen[1] == 0 || iSearchPartsLen[2] != 1 || iSearchPartsLen[3] != 1 ||
        	       iSearchPartsLen[4] == 0 || iSearchPartsLen[5] == 0 || iSearchPartsLen[6] == 0)
        	       return;
       
        		string seeker(sSearchParts[0], iSearchPartsLen[0]+iSearchPartsLen[1]+1);
        		
        		bool bPassive = iSearchPartsLen[0] == 3 && ((uint32_t *)(sSearchParts[0]))[0] == ((uint32_t *)"Hub:")[0];
        
        		// We don't wan't to answer passive searches if we're in passive mode...
        		if(bPassive == true && ClientManager::getInstance()->isActive() == false) {
        			return;
        		}
        		
        		// Filter own searches
        		if(ClientManager::getInstance()->isActive() && bPassive == false) {
                    string sIP = ClientManager::getInstance()->getCachedIp();
                    string sPort = Util::toString(SearchManager::getInstance()->getPort());
                    if(iSearchPartsLen[0] == sIP.size() && strncmp(sSearchParts[0], sIP.c_str(), iSearchPartsLen[0]) == NULL &&
                        iSearchPartsLen[1] == sPort.size() && strcmp(sSearchParts[1], sPort.c_str()) == NULL) {
        				return;
        			}
        		} else {
                    string sNick = getCurrentNick();
        			if(iSearchPartsLen[1] == sNick.size() && strcmp(sSearchParts[1], sNick.c_str()) == 0) {
        				return;
        			}
        		}
        
        		uint64_t tick = GET_TICK();
        		clearFlooders(tick);
        		seekers.push_back(make_pair(seeker, tick));
        		
        		// First, check if it's a flooder
        		for(FloodIter fi = flooders.begin(); fi != flooders.end(); ++fi) {
        			if(fi->first == seeker)
        				return;
        		}
        		int count = 0;
        		for(FloodIter fi = seekers.begin(); fi != seekers.end(); ++fi) {
        			if(fi->first == seeker)
        				count++;
        
        			if(count > 7) {
        				// BFU don't need to know search spammers ;)
        				if(isOp()) {
        					if(bPassive == true) {
                                string sText = string(sSearchParts[1], iSearchPartsLen[1]);
        						fire(ClientListener::SearchFlood(), this, sText);
        					} else {
        						fire(ClientListener::SearchFlood(), this, seeker + STRING(NICK_UNKNOWN));
        					}
        				}				
        				flooders.push_back(make_pair(seeker, tick));
        				return;
        			}
        		}
        		
        		int a;
        		if(sSearchParts[2][0] == 'F') {
        			a = SearchManager::SIZE_DONTCARE;
        		} else {
        			if(sSearchParts[3][0] == 'F') {
        				a = SearchManager::SIZE_ATLEAST;
        			} else {
        				a = SearchManager::SIZE_ATMOST;
        			}
        		}
      
        		int64_t size = _atoi64(sSearchParts[4]);
        		int type = atoi(sSearchParts[5]) - 1;

        		if(bPassive == true) {
                    string sNick = string(sSearchParts[1], iSearchPartsLen[1]);
        			OnlineUser* u = findUser(sNick);
        				
        			if(u == NULL) return;
                        
                    if(!u->getUser()->isSet(User::PASSIVE)) {
        				u->getUser()->setFlag(User::PASSIVE);
        				updated(*u);
        			}
        		}

                string sFile = unescape(string(sSearchParts[6], iSearchPartsLen[6]));
        		fire(ClientListener::NmdcSearch(), this, seeker, a, size, type, sFile, bPassive);

        		return;
            }
            
            // $SR
            if(((uint16_t *)(aLine+2))[0] == ((uint16_t *)"R ")[0]) {
                if(iaLineLen < 5) return;
                
    			SearchManager::getInstance()->onSearchResult(line);
    			return;
            }
            
            // $Supports
           	if(((uint32_t *)(aLine+2))[0] == ((uint32_t *)"uppo")[0] && ((uint16_t *)(aLine+6))[0] == ((uint16_t *)"rt")[0] && aLine[8] == 's') {
        		if(iaLineLen > 10) {
                    char *sSupport = aLine+10;
                    
                    for(int i = 10; i < iaLineLen; i++) {
                        if(aLine[i] == ' ') {
                            aLine[i] = '\0';
                        } else if(i != iaLineLen-1) {
                            continue;
                        } else {
                            i++;
                        }
                        
                        int iDataLen = (aLine+i)-sSupport;
                        
                		switch(sSupport[0]) {
                            case 'U':
                        		if(iDataLen == 11 && ((uint64_t *)(sSupport+1))[0] == ((uint64_t *)"serComma")[0] && ((uint16_t *)(sSupport+9))[0] == ((uint16_t *)"nd")[0]) {
                        			setFlag(SUPPORTS_USERCOMMAND);
                        		} else if(iDataLen == 7 && ((uint32_t *)(sSupport+1))[0] == ((uint32_t *)"serI")[0] && ((uint16_t *)(sSupport+5))[0] == ((uint16_t *)"P2")[0]) {
                        			setFlag(SUPPORTS_USERIP2);
                                }
                                break;
                            case 'N':
                        		if(iDataLen == 9 && ((uint64_t *)(sSupport+1))[0] == ((uint64_t *)"oGetINFO")[0]) {
                        			setFlag(SUPPORTS_NOGETINFO);
                                }
                                break;
                            case 'Q':
                        		if(iDataLen == 9 && ((uint64_t *)(sSupport+1))[0] == ((uint64_t *)"uickList")[0]) {
                        			if(state == STATE_HELLO) {
                        				state = STATE_MYINFO;
                        				updateCounts(false);
                        				myInfo();
                        				getNickList();
                        				setFlag(SUPPORTS_QUICKLIST);
                        			}
                        		}
                        		break;
                        	default:
                                dcdebug("NmdcHub::onLine Unknown supports %s\n", aLine);
                                break;
                        }
                        sSupport = aLine+i+1;
                    }
                }
                
        		if(isSet(SUPPORTS_QUICKLIST) == false) {
        			OnlineUser& ou = getUser(getCurrentNick());
        			validateNick(ou.getIdentity().getNick());
        		}
        		return;
            }
            
            break;
    	case 'M':
        	// $MyINFO $ALL  $ $$$$
        	if(((uint64_t *)(aLine+2))[0] == ((uint64_t *)"yINFO $A")[0] && ((uint16_t *)(aLine+10))[0] == ((uint16_t *)"LL")[0]) {
                if(iaLineLen < 21)
                    return;
              
                char *sMyINFOParts[] = { NULL, NULL, NULL, NULL, NULL, NULL };
                unsigned short iMyINFOPartsLen[] = { 0, 0, 0, 0, 0, 0 };
                
                unsigned char cPart = 0;
                
                sMyINFOParts[cPart] = aLine+13; // Nick
                
                for(int i = 13; i < iaLineLen; i++) {
                    if(aLine[i] == '$') {
                        aLine[i] = '\0';
                        iMyINFOPartsLen[cPart] = (unsigned short)((aLine+i)-sMyINFOParts[cPart]);
                        
                        // are we on end of myinfo ???
                        if(cPart == 5)
                            break;
                            
                        cPart++;
                        sMyINFOParts[cPart] = aLine+i+1;
                    } else if(cPart == 0 && aLine[i] == ' ') {
                        aLine[i] = '\0';
                        iMyINFOPartsLen[cPart] = (unsigned short)((aLine+i)-sMyINFOParts[cPart]);
                        cPart++;
                        sMyINFOParts[cPart] = aLine+i+1;
                    }
                }

                // check if we have nick...
        	    if(iMyINFOPartsLen[0] == 0)
        	         return;
               
                string sNick = string(sMyINFOParts[0], iMyINFOPartsLen[0]);
        		OnlineUser& u = getUser(sNick);
        		
        		// nick case sensitivity fix
        		if(strcmp(u.getIdentity().getNick().c_str(), sMyINFOParts[0]) != NULL) {
                    u.getIdentity().set("NI", sNick);
                }
                       
                // clear all tag parts
        		u.getIdentity().set("CL", Util::emptyString);
        		u.getIdentity().set("VE", Util::emptyString);
        		u.getIdentity().set("MD", Util::emptyString);
        		u.getIdentity().set("HT", Util::emptyString);
        		u.getIdentity().set("HN", Util::emptyString);
				u.getIdentity().set("HR", Util::emptyString);
				u.getIdentity().set("HO", Util::emptyString);
				u.getIdentity().set("SL", Util::emptyString);
        		u.getIdentity().set("UL", Util::emptyString);
        		
                // check if we have description + tag
                if(iMyINFOPartsLen[1] != 0) {
                    if(sMyINFOParts[1][iMyINFOPartsLen[1]-1] == '>') {
                        char *sTag;
        				if((sTag = strrchr(sMyINFOParts[1], '<')) != NULL) {
                            string sTA = string(sTag, iMyINFOPartsLen[1]-(sTag-sMyINFOParts[1]));
        					u.getIdentity().set("TA", sTA);
        					// Hm, we have something...disassemble it...
							updateFromTag(u.getIdentity());
							if(sTag-sMyINFOParts[1] != 0) {
                                sTag[0] = '\0';
                                string sDescr = string(sMyINFOParts[1], sTag-sMyINFOParts[1]);
                                u.getIdentity().setDescription(sDescr);
                            } else {
                                u.getIdentity().setDescription(Util::emptyString);
                            }
        				} else {
                            string sDescr = string(sMyINFOParts[1], iMyINFOPartsLen[1]);
							u.getIdentity().setDescription(sDescr);
                            u.getIdentity().set("TA", Util::emptyString);
                        }
        			} else {
                        string sDescr = string(sMyINFOParts[1], iMyINFOPartsLen[1]);
						u.getIdentity().setDescription(sDescr);
						u.getIdentity().set("TA", Util::emptyString);
					}
                } else {
                    u.getIdentity().setDescription(Util::emptyString);
                    u.getIdentity().set("TA", Util::emptyString);
                }
                
                // check if we have 'space' before connection '$'
                if(iMyINFOPartsLen[2] != 0 && sMyINFOParts[2][0] != ' ') {
                    if(sMyINFOParts[2][0] == 'A') {
            			u.getIdentity().getUser()->unsetFlag(User::PASSIVE);
            		} else {
            			u.getIdentity().getUser()->setFlag(User::PASSIVE);
                    }
                }
                
                // check if we have connection
                if(iMyINFOPartsLen[3] != 0) {
                    char status = sMyINFOParts[3][iMyINFOPartsLen[3]-1];
            		sMyINFOParts[3][iMyINFOPartsLen[3]-1] = '\0';
            		if(iMyINFOPartsLen[3] != 1) {
                        string sConn = string(sMyINFOParts[3], iMyINFOPartsLen[3]-1);
                        u.getIdentity().setConnection(sConn);
                        
               			// set connection type
                		switch(sMyINFOParts[3][0]) {
                            case '0':
                                if(sMyINFOParts[3][1] == '.') {
                                    if(((uint32_t *)(sMyINFOParts[3]+2))[0] == ((uint32_t *)"005\0")[0] || 
                                        (sMyINFOParts[3][4] == '\0' && (((uint16_t *)(sMyINFOParts[3]+2))[0] == ((uint16_t *)"01")[0] || 
                                        ((uint16_t *)(sMyINFOParts[3]+2))[0] == ((uint16_t *)"02")[0] || 
                                        ((uint16_t *)(sMyINFOParts[3]+2))[0] == ((uint16_t *)"05")[0]))) {
                                        u.getIdentity().setcType(1);
                                        break;
                                    } else if(((uint16_t *)(sMyINFOParts[3]+2))[0] == ((uint16_t *)"1\0")[0] || 
                                        ((uint16_t *)(sMyINFOParts[3]+2))[0] == ((uint16_t *)"2\0")[0] || 
                                        ((uint16_t *)(sMyINFOParts[3]+2))[0] == ((uint16_t *)"5\0")[0]) {
                                        u.getIdentity().setcType(5);
                                        break;
                                    }
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case '1':
                                if(sMyINFOParts[3][1] == '\0') {
                                    u.getIdentity().setcType(5);
                                    break;
                                } else if(((uint16_t *)(sMyINFOParts[3]+1))[0] == ((uint16_t *)"0\0")[0]) {
                                    u.getIdentity().setcType(6);
                                    break;
                                } else if(sMyINFOParts[3][3] == '\0' && ((uint16_t *)(sMyINFOParts[3]+1))[0] == ((uint16_t *)"00")[0]) {
                                    u.getIdentity().setcType(7);
                                    break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case '2':
                                if(sMyINFOParts[3][1] == '\0') {
                                    u.getIdentity().setcType(6);
                                    break;
                                } else if(((uint16_t *)(sMyINFOParts[3]+1))[0] == ((uint16_t *)"0\0")[0]) {
                                    u.getIdentity().setcType(7);
                                    break;
                                } else if(((uint64_t *)(sMyINFOParts[3]+1))[0] == ((uint64_t *)"8.8Kbps\0")[0]) {
                                    u.getIdentity().setcType(1);
                                    break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case '3':
                    			if(((uint64_t *)(sMyINFOParts[3]+1))[0] == ((uint64_t *)"3.6Kbps\0")[0]) {
                                    u.getIdentity().setcType(1);
                                    break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case '5':
                    			if(sMyINFOParts[3][1] == '\0') {
                                    u.getIdentity().setcType(6);
                                    break;
                                } else if(((uint16_t *)(sMyINFOParts[3]+1))[0] == ((uint16_t *)"0\0")[0]) {
                                    u.getIdentity().setcType(7);
                                    break;
                                } else if(((uint32_t *)(sMyINFOParts[3]+1))[0] == ((uint32_t *)"6Kbp")[0] && 
                                    ((uint16_t *)(sMyINFOParts[3]+5))[0] == ((uint16_t *)"s\0")[0]) {
                                    u.getIdentity().setcType(1);
                                    break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'M':
                                if(sMyINFOParts[3][5] == '\0' && ((uint32_t *)(sMyINFOParts[3]+1))[0] == ((uint32_t *)"odem")[0]) {
                    				u.getIdentity().setcType(1);
                    				break;
                                } else if(sMyINFOParts[3][9] == '\0' && ((uint64_t *)(sMyINFOParts[3]+1))[0] == ((uint64_t *)"icrowave")[0]) {
                    				u.getIdentity().setcType(3);
                    				break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'I':
                        		if(((uint32_t *)(sMyINFOParts[3]+1))[0] == ((uint32_t *)"SDN\0")[0]) {
                    				u.getIdentity().setcType(2);
                    				break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'S':
                        		if(sMyINFOParts[3][9] == '\0' && ((uint64_t *)(sMyINFOParts[3]+1))[0] == ((uint64_t *)"atellite")[0]) {
                                    u.getIdentity().setcType(3);
                                    break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'W':
                                if(((uint64_t *)(sMyINFOParts[3]+1))[0] == ((uint64_t *)"ireless\0")[0]) {
                    				u.getIdentity().setcType(3);
                    				break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'D':
                        		if(sMyINFOParts[3][3] == '\0' && ((uint16_t *)(sMyINFOParts[3]+1))[0] == ((uint16_t *)"SL")[0]) {
                    				u.getIdentity().setcType(4);
                    				break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'C':
                        		if(sMyINFOParts[3][5] == '\0' && ((uint32_t *)(sMyINFOParts[3]+1))[0] == ((uint32_t *)"able")[0]) {
                    				u.getIdentity().setcType(5);
                    				break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'L':
                                if(sMyINFOParts[3][7] == '\0') {
                            		if(((uint32_t *)(sMyINFOParts[3]+1))[0] == ((uint32_t *)"AN(T")[0] && 
                                        ((uint16_t *)(sMyINFOParts[3]+5))[0] == ((uint16_t *)"1)")[0]) {
                                        u.getIdentity().setcType(6);
                                        break;
                                    } if(((uint32_t *)(sMyINFOParts[3]+1))[0] == ((uint32_t *)"AN(T")[0] && 
                                        ((uint16_t *)(sMyINFOParts[3]+5))[0] == ((uint16_t *)"3)")[0]) {
                                        u.getIdentity().setcType(7);
                                        break;
                                    }
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'N':
                                if(((uint64_t *)(sMyINFOParts[3]+1))[0] == ((uint64_t *)"etlimite")[0] && sMyINFOParts[3][9] == 'r') {
                                    u.getIdentity().setcType(8);
                                    break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            default:
                                // no known connection type found, set unknown
                                u.getIdentity().setcType(11);
                                break;
                        }
                    } else {
                        u.getIdentity().setConnection(Util::emptyString);
                        u.getIdentity().setcType(11);
                    }
                    
                    // set status based on magich byte
                    if((status & 0x8) == 0x8) {
               			u.getIdentity().setFlag(Identity::FIREBALL);
                    } else {
                        u.getIdentity().unsetFlag(Identity::FIREBALL);
                    }

                    if((status & 0x4) == 0x4) {
               			u.getIdentity().setFlag(Identity::SERVER);
                    } else {
                        u.getIdentity().unsetFlag(Identity::SERVER);
                    }

                    if((status & 0x2) == 0x2) {
              			u.getIdentity().setFlag(Identity::AWAY);
                    } else {
                        u.getIdentity().unsetFlag(Identity::AWAY);
                    }
                } else {
                    u.getIdentity().setConnection(Util::emptyString);
                    u.getIdentity().setcType(11);
                    u.getIdentity().unsetFlag(Identity::FIREBALL);
                    u.getIdentity().unsetFlag(Identity::SERVER);
                    u.getIdentity().unsetFlag(Identity::AWAY);
                }
                
                // check if we have email
                if(iMyINFOPartsLen[4] != 0) {
                    string sEmail = string(sMyINFOParts[4], iMyINFOPartsLen[4]);
                    u.getIdentity().setEmail(sEmail);
                } else {
                    u.getIdentity().setEmail(Util::emptyString);
                }
                
                // check if we have sharesize
                if(iMyINFOPartsLen[5] != 0) {
                    string sShare = string(sMyINFOParts[5], iMyINFOPartsLen[5]);
                    u.getIdentity().setBytesShared(sShare);
                } else {
                    u.getIdentity().setBytesShared("0");
                }
                             		
        		if(state == STATE_MYINFO) {
					if(u.getUser() == getMyIdentity().getUser()) {
        				state = STATE_LOGGEDIN;
						setMyIdentity(u.getIdentity());
        				updateCounts(false);
        				u.getUser()->setFlag(User::DCPLUSPLUS);
        				if(SETTING(CONNECTION_TYPE) != SettingsManager::CONNECTION_ACTIVE)
        					u.getUser()->setFlag(User::PASSIVE);
        				else
        					u.getUser()->unsetFlag(User::PASSIVE);
        			}
        		}
				fire(ClientListener::UserUpdated(), this, u);
        		return;
            }

            // $MultiConnectToMe
    		if(((uint64_t *)(aLine+2))[0] == ((uint64_t *)"ultiConn")[0] && ((uint64_t *)(aLine+10))[0] == ((uint64_t *)"ectToMe ")[0]) {
                if(state != STATE_LOGGEDIN || iaLineLen < 21)
                    return;

                char *sMCTMParts[] = { NULL, NULL, NULL };
                unsigned short iMCTMPartsLen[] = { 0, 0, 0 };
                
                bool bFirst = true;
                
                sMCTMParts[0] = aLine+18; // Nick
                
                for(int i = 18; i < iaLineLen; i++) {
                    if(bFirst == true) {
                        if(aLine[i] == ' ') {
                            aLine[i] = '\0';
                            iMCTMPartsLen[0] = (unsigned short)((aLine+i)-sMCTMParts[0]);
                            sMCTMParts[1] = aLine+i+1;
                            bFirst = false;
                        }
                    } else if(aLine[i] == ':') {
                        aLine[i] = '\0';
                        iMCTMPartsLen[1] = (unsigned short)((aLine+i)-sMCTMParts[1]);
                        sMCTMParts[2] = aLine+i+1;
                        iMCTMPartsLen[2] = (unsigned short)(iaLineLen-i-1);
                        break;
                    }
                }
                
                // check if we have all we need
                if(iMCTMPartsLen[0] == 0 || iMCTMPartsLen[1] == 0 || iMCTMPartsLen[2] == 0)
                    return;

                string sNick = getCurrentNick();
        		if(iMCTMPartsLen[0] != sNick.size() || _stricmp(sMCTMParts[0], sNick.c_str()) != 0) // Check nick... is MCTM really for me ? ;o)
        			return;

                string sIP = string(sMCTMParts[1], iMCTMPartsLen[1]);
        		ConnectionManager::getInstance()->nmdcConnect(sIP, (unsigned short)atoi(sMCTMParts[2]), 
                    getCurrentNick(), getHubUrl());
        		return;
            }
            
            // $MultiSearch
        	if(((uint64_t *)(aLine+2))[0] == ((uint64_t *)"ultiSear")[0] && ((uint16_t *)(aLine+10))[0] == ((uint16_t *)"ch")[0] && aLine[12] == ' ') {   		
        	    if(state != STATE_LOGGEDIN || iaLineLen < 24)
                    return;
       	    
        	    char *sSearchParts[] = { NULL, NULL, NULL, NULL, NULL, NULL, NULL };
                unsigned short iSearchPartsLen[] = { 0, 0, 0, 0, 0, 0, 0 };
                
                unsigned char cPart = 0;
                
                sSearchParts[cPart] = aLine+13; // IP
                
        	    for(int i = 13; i < iaLineLen; i++) {
                    switch(aLine[i]) {
                        case ':':
                            if(cPart == 0) {
                                iSearchPartsLen[cPart] = (unsigned short)((aLine+i)-sSearchParts[cPart]);
                                cPart++;
                                sSearchParts[cPart] = aLine+i+1;
                            }
                            break;
                        case ' ':
                            if(cPart == 1) {
                                aLine[i] = '\0';
                                iSearchPartsLen[cPart] = (unsigned short)((aLine+i)-sSearchParts[cPart]);
                                cPart++;
                                sSearchParts[cPart] = aLine+i+1;
                            }
                            break;
                        case '?':
                            if(cPart > 1) {
                                aLine[i] = '\0';
                                iSearchPartsLen[cPart] = (unsigned short)((aLine+i)-sSearchParts[cPart]);
                                
                                // are we on last '?' ?
                                if(cPart == 5) {
                                    sSearchParts[6] = aLine+i+1;
                                    iSearchPartsLen[6] = (unsigned short)(iaLineLen-i-1);
                                    i = iaLineLen;
                                    break;
                                }
                                    
                                cPart++;
                                sSearchParts[cPart] = aLine+i+1;
                            }
                            break;
                    }
                }
       	    
       	        // check if we have all we need
        	    if(iSearchPartsLen[0] == 0 || iSearchPartsLen[1] == 0 || iSearchPartsLen[2] != 1 || iSearchPartsLen[3] != 1 ||
        	       iSearchPartsLen[4] == 0 || iSearchPartsLen[5] == 0 || iSearchPartsLen[6] == 0)
        	       return;
       
        		string seeker(sSearchParts[0], iSearchPartsLen[0]+iSearchPartsLen[1]+1);
        		
        		// Filter own searches
        		if(ClientManager::getInstance()->isActive()) {
                    string sIP = ClientManager::getInstance()->getCachedIp();
                    string sPort = Util::toString(SearchManager::getInstance()->getPort());
                    if(iSearchPartsLen[0] == sIP.size() && strncmp(sSearchParts[0], sIP.c_str(), iSearchPartsLen[0]) == NULL &&
                        iSearchPartsLen[1] == sPort.size() && strcmp(sSearchParts[1], sPort.c_str()) == NULL) {
        				return;
        			}
        		}
        
        		uint64_t tick = GET_TICK();
        		clearFlooders(tick);
        		seekers.push_back(make_pair(seeker, tick));
        		// First, check if it's a flooder
        		for(FloodIter fi = flooders.begin(); fi != flooders.end(); ++fi) {
        			if(fi->first == seeker)
        				return;
        		}
        		int count = 0;
        		for(FloodIter fi = seekers.begin(); fi != seekers.end(); ++fi) {
        			if(fi->first == seeker)
        				count++;
        
        			if(count > 7) {
        				// BFU don't need to know search spammers ;)
        				if(isOp()) {
        					fire(ClientListener::SearchFlood(), this, seeker + STRING(NICK_UNKNOWN));
        				}				
        				flooders.push_back(make_pair(seeker, tick));
        				return;
        			}
        		}
        		
        		int a;
        		if(sSearchParts[2][0] == 'F') {
        			a = SearchManager::SIZE_DONTCARE;
        		} else {
        			if(sSearchParts[3][0] == 'F') {
        				a = SearchManager::SIZE_ATLEAST;
        			} else {
        				a = SearchManager::SIZE_ATMOST;
        			}
        		}
      
        		int64_t size = _atoi64(sSearchParts[4]);
        		int type = atoi(sSearchParts[5]) - 1;

                string sFile = unescape(string(sSearchParts[6], iSearchPartsLen[6]));
        		fire(ClientListener::NmdcSearch(), this, seeker, a, size, type, sFile, false);
        		
        		return;
            }
            
            break;
        case 'Q':
        	// $Quit
    		if(((uint32_t *)(aLine+2))[0] == ((uint32_t *)"uit ")[0]) {
                if(iaLineLen < 7)
                    return;
                
                string sNick = string(aLine+6, iaLineLen-6);
        		OnlineUser* u = findUser(sNick);
        		
				if(u == NULL)
                    return;  
				      		
        		fire(ClientListener::UserRemoved(), this, *u);
        		putUser(sNick);

        		return;
            }
            
            break;
        case 'C':
    		// $ConnectToMe
    		if(((uint64_t *)(aLine+2))[0] == ((uint64_t *)"onnectTo")[0] && ((uint16_t *)(aLine+10))[0] == ((uint16_t *)"Me")[0] && aLine[12] == ' ') {
                if(state != STATE_LOGGEDIN || iaLineLen < 16)
                    return;

                char *sCTMParts[] = { NULL, NULL, NULL };
                unsigned short iCTMPartsLen[] = { 0, 0, 0 };
                
                bool bFirst = true;
                
                sCTMParts[0] = aLine+13; // Nick
                
                for(int i = 13; i < iaLineLen; i++) {
                    if(bFirst == true) {
                        if(aLine[i] == ' ') {
                            aLine[i] = '\0';
                            iCTMPartsLen[0] = (unsigned short)((aLine+i)-sCTMParts[0]);
                            sCTMParts[1] = aLine+i+1;
                            bFirst = false;
                        }
                    } else if(aLine[i] == ':') {
                        aLine[i] = '\0';
                        iCTMPartsLen[1] = (unsigned short)((aLine+i)-sCTMParts[1]);
                        sCTMParts[2] = aLine+i+1;
                        iCTMPartsLen[2] = (unsigned short)(iaLineLen-i-1);
                        break;
                    }
                }
                
                // check if we have all we need
                if(iCTMPartsLen[0] == 0 || iCTMPartsLen[1] == 0 || iCTMPartsLen[2] == 0)
                    return;

                string sNick = getCurrentNick();
        		if(iCTMPartsLen[0] != sNick.size() || _stricmp(sCTMParts[0], sNick.c_str()) != 0) // Check nick... is CTM really for me ? ;o)
        			return;

                string sIP = string(sCTMParts[1], iCTMPartsLen[1]);
        		ConnectionManager::getInstance()->nmdcConnect(sIP, (uint16_t)atoi(sCTMParts[2]), 
                    getCurrentNick(), getHubUrl());

        		return;
            }
            
            break;
        case 'R':
        	// $RevConnectToMe
        	if(((uint64_t *)(aLine+2))[0] == ((uint64_t *)"evConnec")[0] && ((uint32_t *)(aLine+10))[0] == ((uint32_t *)"tToM")[0] && 
                ((uint16_t *)(aLine+14))[0] == ((uint16_t *)"e ")[0]) {
        		if(state != STATE_LOGGEDIN || iaLineLen < 19)
                    return;

				char *sTemp = strchr(aLine+16, ' ');
				if(sTemp == NULL || (sTemp-aLine)-16 == 0 || iaLineLen-(sTemp-aLine)-1 == 0)
					return;
					
        		string sMyNick = getCurrentNick();
        		if(size_t(iaLineLen-(sTemp-aLine)-1) != sMyNick.size() || _stricmp(sTemp+1, sMyNick.c_str()) != 0) // Check nick... is RCTM really for me ? ;o)
        			return;

        		string sNick = string(aLine+16, (sTemp-aLine)-16);
        		OnlineUser* u = findUser(sNick);
        		
        		if(u == NULL)
                    return;
        
        		if(ClientManager::getInstance()->isActive()) {
        			connectToMe(*u);
        			if(u->getUser()->isSet(User::PASSIVE) == false) {
        				u->getUser()->setFlag(User::PASSIVE);
        				updated(*u);
                    }
        		} else {
        			// Notify the user that we're passive too...
        			if(u->getUser()->isSet(User::PASSIVE) == false) {
        				if(u->getIdentity().isSet(Identity::RCTM_SND) == false) {
        					revConnectToMe(*u);
        					u->getIdentity().setFlag(Identity::RCTM_SND);
        				}
        				u->getUser()->setFlag(User::PASSIVE);
        				updated(*u);
                    }
        		}

        		return;
            }
            
            break;
        case 'H':
        	// $HubName
        	if(((uint32_t *)(aLine+2))[0] == ((uint32_t *)"ubNa")[0] && ((uint16_t *)(aLine+6))[0] == ((uint16_t *)"me")[0] && aLine[8] == ' ') {
                if(iaLineLen < 10)
                    return;
                
                char *sTmp = strstr(aLine+9, " - ");
        		if(sTmp != NULL && iaLineLen-3-(sTmp-aLine) != 0) {
                    string sHubName = unescape(string(aLine+9, sTmp-aLine-9));
                    getHubIdentity().setNick(sHubName);
                    string sHubTopic = unescape(string(sTmp+3, iaLineLen-3-(sTmp-aLine)));
                    getHubIdentity().setDescription(sHubTopic);
        		} else {
                    string sHubName = unescape(string(aLine+9, iaLineLen-9));
					getHubIdentity().setNick(sHubName);
					getHubIdentity().setDescription(Util::emptyString);	
        		}    			
        		fire(ClientListener::HubUpdated(), this);
        		return;
            }
            
            // $Hello
           	if(((uint32_t *)(aLine+2))[0] == ((uint32_t *)"ello")[0] && aLine[6] == ' ') {
        		if(isSet(SUPPORTS_QUICKLIST) == true || iaLineLen < 8)
                    return;
      
        		string nick(aLine+7, iaLineLen-7);
        		OnlineUser& u = getUser(nick);
        		u.getIdentity().set("NI", nick);
        		
        		if(state == STATE_HELLO) {
            		if(u.getUser() == getMyIdentity().getUser()) {
                        state = STATE_LOGGEDIN;
        			    updateCounts(false);
            			u.getUser()->setFlag(User::DCPLUSPLUS);
            			if(ClientManager::getInstance()->isActive())
            				u.getUser()->unsetFlag(User::PASSIVE);
            			else
            				u.getUser()->setFlag(User::PASSIVE);
            		}
        
        			version();
        			if(ShittYnHub == true || getStealth() == true) {
        				getNickList();
        				myInfo();
                    } else {
        				myInfo();
        				getNickList();
                    }
        		}
				fire(ClientListener::UserUpdated(), this, u);
        		return;
            }
            
            // $HubIsFull
           	if(iaLineLen == 10 && ((uint64_t *)(aLine+2))[0] == ((uint64_t *)"ubIsFull")[0]) {
				fire(ClientListener::HubFull(), this);
    			return;
        	}
        	
        	break;
        case 'U':
        	// $UserCommand
        	if(((uint64_t *)(aLine+2))[0] == ((uint64_t *)"serComma")[0] && ((uint16_t *)(aLine+10))[0] == ((uint16_t *)"nd")[0] && aLine[12] == ' ') {
                if(iaLineLen < 16)
                    return;
                
                char *sUsrCmdParts[] = { NULL, NULL, NULL, NULL };
                unsigned short iUsrCmdPartsLen[] = { 0, 0, 0, 0 };
               
                unsigned char cPart = 0;
                sUsrCmdParts[cPart] = aLine+13; // type
                
                for(int i = 13; i < iaLineLen; i++) {
                    switch(cPart) {
                        case 0:
                        case 1:
                            if(aLine[i] == ' ') {
                                aLine[i] = '\0';
                            } else if(i != iaLineLen-1) {
                                continue;
                            } else {
                                i++;
                            }
                            iUsrCmdPartsLen[cPart] = (unsigned short)((aLine+i)-sUsrCmdParts[cPart]);
                            cPart++;
                            sUsrCmdParts[cPart] = aLine+i+1;
                            break;
                        case 2:
                            if(aLine[i] == '$') {
                                aLine[i] = '\0';
                                iUsrCmdPartsLen[2] = (unsigned short)((aLine+i)-sUsrCmdParts[2]);
                                sUsrCmdParts[3] = aLine+i+1;
                                iUsrCmdPartsLen[3] = (unsigned short)(iaLineLen-i-1);
                                i = iaLineLen;
                            }
                            break;
                    }
                }
               
                // check if we have two needed numbers...
                if(iUsrCmdPartsLen[0] == 0 || iUsrCmdPartsLen[1] == 0)
                    return;

        		int type = atoi(sUsrCmdParts[0]);
        		int ctx = atoi(sUsrCmdParts[1]);;
        
        		switch(type) {
                    case UserCommand::TYPE_SEPARATOR:
                    case UserCommand::TYPE_CLEAR:
        			    fire(ClientListener::UserCommand(), this, type, ctx, Util::emptyString, Util::emptyString);
        			    break;
            		case UserCommand::TYPE_RAW:
					case UserCommand::TYPE_RAW_ONCE: {
                        // check if we have two needed strings...
                        if(iUsrCmdPartsLen[2] == 0 || iUsrCmdPartsLen[3] == 0)
                            return;

                        string sCmdName = unescape(string(sUsrCmdParts[2], iUsrCmdPartsLen[2]));
                        string sCmdStr = unescape(string(sUsrCmdParts[3], iUsrCmdPartsLen[3]));
        			    fire(ClientListener::UserCommand(), this, type, ctx, sCmdName, sCmdStr);
        			    break;
					}
        			default:
                        // ?!?
                        break;
        		}
        		return;
            }
            
            // $UserIP
           	if(((uint32_t *)(aLine+2))[0] == ((uint32_t *)"serI")[0] && ((uint16_t *)(aLine+6))[0] == ((uint16_t *)"P ")[0]) {
                if(iaLineLen < 17)
                    return;
                
        		OnlineUser::List v;
        		
        		char *sNick = aLine+8; // first nick
        		char *sIp = NULL;
        		bool bSpace = true;
        		
        		for(int i = 8; i < iaLineLen; i++) {
                    if(bSpace == true) {
                        if(aLine[i] == ' ') {
                            aLine[i] = '\0';    
                            sIp = aLine+i+1;    
                            bSpace = false;    
                        }        
                    } else {
                        if(aLine[i] == '$') {
                            aLine[i] = '\0';
                        } else if(i != iaLineLen-1) {
                            continue;
                        } else {
                            i++;
                        }

                        string sName = string(sNick, (sIp-sNick)-1);
                        OnlineUser* u = findUser(sName);
                        if(u == NULL)
                            continue;

                        string sIP = string(sIp, (aLine+i)-sIp);
                        u->getIdentity().setIp(sIP);
                        v.push_back(u);
                        
                        sNick = aLine+i+2;
                        i++;
                        bSpace = true;    
                    }                  
                }

				fire(ClientListener::UsersUpdated(), this, v);
        		return;
            }
            
            break;
        case 'L':
        	// $Lock
        	if(((uint32_t *)(aLine+2))[0] == ((uint32_t *)"ock ")[0]) {
        		if(state != STATE_LOCK || l.size() < 7) return;
        		
        		state = STATE_HELLO;
        		string lck;
        		
        		char *sTmp = strstr((char *)l.c_str()+6, " Pk=");
        		if(sTmp != NULL) {
        			lck = l.substr(6, sTmp-l.c_str()-6);
        			if(_stricmp(sTmp+4, "YnHub") == 0) {
        				ShittYnHub = true;
        			} else if(strcmp(sTmp+4, "PtokaX") == 0) {
        				PtokaX = true;
        			}
                } else {
                    lck = l;
                }
                
        		if(CryptoManager::getInstance()->isExtended(lck)) {
        			StringList feat;
        			feat.push_back("UserCommand"); feat.push_back("NoGetINFO");
            		feat.push_back("NoHello"); feat.push_back("UserIP2");
        			if(getStealth() == false) {
            			feat.push_back("QuickList");
            			feat.push_back("ZPipe");
                    } else {
        				feat.push_back("TTHSearch");
        				feat.push_back("ZPipe");
                    }

        			supports(feat);
					if(PtokaX == false || getStealth() == true)
                        key(CryptoManager::getInstance()->makeKey(lck));
        		} else {
        			key(CryptoManager::getInstance()->makeKey(lck));
					OnlineUser& ou = getUser(getCurrentNick());
					validateNick(ou.getIdentity().getNick());
        		}
                return;
    		}

    		// $LogedIn
    		if(((uint32_t *)(aLine+2))[0] == ((uint32_t *)"oged")[0] && ((uint16_t *)(aLine+6))[0] == ((uint16_t *)"In")[0]) {
    			// nothing to do here...
    			return;
            }
            
            break;
    	case 'F':
            // $ForceMove
        	if(((uint64_t *)(aLine+2))[0] == ((uint64_t *)"orceMove")[0] && aLine[10] == ' ') {
                if(iaLineLen < 12)
                    return;

        		socket->disconnect(false);
        		string sFrcMv = string(aLine+11, iaLineLen-11);
        		fire(ClientListener::Redirect(), this, sFrcMv);
        		return;
            }
            
            break;
        case 'V':
        	// $ValidateDenide
        	if(((uint64_t *)(aLine+2))[0] == ((uint64_t *)"alidateD")[0] && ((uint32_t *)(aLine+10))[0] == ((uint32_t *)"enid")[0] && aLine[15] == 'e') {
        		socket->disconnect(false);
				fire(ClientListener::NickTaken(), this);
        		return;
            }
            
            break;
        case 'N':
        	// $NickList
        	if(((uint64_t *)(aLine+2))[0] == ((uint64_t *)"ickList ")[0]) {
                if(iaLineLen < 11)
                    return;
                    
                OnlineUser::List v;
                char *sNick = aLine+10; // first nick

                for(int i = 10; i < iaLineLen; i++) {
                    if(aLine[i] == '$') {
                        string sUserNick = string(sNick, (aLine+i)-sNick);			
    					v.push_back(&getUser(sUserNick));
                        sNick = aLine+i+2;
                        i++;
                    }
                }

        		if(isSet(SUPPORTS_NOGETINFO) == false) {
        			string tmp;
        			// Let's assume 10 characters per nick...
        			tmp.reserve(v.size() * (11 + 10 + getCurrentNick().length())); 
        			string n = ' ' +  fromUtf8(getCurrentNick()) + '|';
        			for(OnlineUser::List::const_iterator i = v.begin(); i != v.end(); ++i) {
        				tmp += "$GetINFO ";
        					tmp += fromUtf8((*i)->getIdentity().getNick());
        					tmp += n;
        			}
        			if(!tmp.empty()) {
        				send(tmp);
        			}
        		}
				fire(ClientListener::UsersUpdated(), this, v);
        		return;
            }
            
            break;
        case 'O':
        	// $OpList
        	if(((uint32_t *)(aLine+2))[0] == ((uint32_t *)"pLis")[0] && ((uint16_t *)(aLine+6))[0] == ((uint16_t *)"t ")[0]) {
                if(iaLineLen < 9)
                    return;
                
                OnlineUser::List v;
                char *sOp = aLine+8; // first OP nick

                for(int i = 8; i < iaLineLen; i++) {
                    if(aLine[i] == '$') {
                        string sOpNick = string(sOp, (aLine+i)-sOp);                      
                        OnlineUser& ou = getUser(sOpNick);
    					ou.getIdentity().setOp(true);
    					if(isOp() == false && sOpNick == getCurrentNick()) {
    						setMyIdentity(ou.getIdentity());
    					}
    					v.push_back(&ou);
                        sOp = aLine+i+2;
                        i++;
                    }
                }
       		
        		fire(ClientListener::UsersUpdated(), this, v);
        		updateCounts(false);
        		if(bFirstOpList == true) {
        			if(getMyIdentity().isRegistered() == true) {
        				myInfo();
                    }
        			bFirstOpList = false;
        		}
        		return;
            }
            
            break;
        case 'T':
        	// $To:
        	if(((uint16_t *)(aLine+2))[0] == ((uint16_t *)"o:")[0] && aLine[4] == ' ') {
                if(iaLineLen < 17)
                    return;
                    
                char *sToParts[] = { NULL, NULL, NULL, NULL };
                unsigned short iToPartsLen[] = { 0, 0, 0, 0 };
                
                unsigned char cPart = 0;
                
                sToParts[cPart] = aLine+5; // Nick
                
                for(int i = 5; i < iaLineLen; i++) {
                    if(aLine[i] == ' ') {
                        aLine[i] = '\0';
                        iToPartsLen[cPart] = (unsigned short)((aLine+i)-sToParts[cPart]);
                        
                        // are we on last space ?
                        if(cPart == 2) {
                            sToParts[3] = aLine+i+2;
                            iToPartsLen[3] = (unsigned short)(iaLineLen-i-2);
                            break;
                        }
                                
                        cPart++;
                        sToParts[cPart] = aLine+i+1;
                    }
                }
               
                // check if we have all we need
                if(iToPartsLen[0] == 0 || iToPartsLen[1] != 5 || iToPartsLen[2] == 0 || iToPartsLen[3] == 0)
                    return;
   
                bool bNormal = true;
                OnlineUser* by = findUserInChat(sToParts[3], bNormal);

                string sFromNick = string(sToParts[2], iToPartsLen[2]);
				OnlineUser* from = findUser(sFromNick);
				string sText = unescape(bNormal ? string(sToParts[3], iToPartsLen[3]) : string(sToParts[3]+4, iToPartsLen[3]-4));
				if(from == NULL ) {
					fire(ClientListener::PrivateMessageNick(), this, sFromNick, *by, sText);
                } else {	
					fire(ClientListener::PrivateMessage(), this, *from, *by, sText);
        		}
        		return;
            }
            
            break;
        case 'G':
        	// $GetPass
        	if(iaLineLen == 8 && ((uint32_t *)(aLine+2))[0] == ((uint32_t *)"etPa")[0] && ((uint16_t *)(aLine+6))[0] == ((uint16_t *)"ss")[0]) {
				OnlineUser& ou = getUser(getCurrentNick());
				ou.getIdentity().setFlag(Identity::REG);
				setMyIdentity(ou.getIdentity());
        		fire(ClientListener::GetPassword(), this);
        		return;
            }
            
            break;
        case 'B':
        	// $BadPass
        	if(iaLineLen == 8 && ((uint32_t *)(aLine+2))[0] == ((uint32_t *)"adPa")[0] && ((uint16_t *)(aLine+6))[0] == ((uint16_t *)"ss")[0]) {
        		setPassword(Util::emptyString);
        		return;
            }
            
            break;
    	default:
		  break;
	}
	
	dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
}

string NmdcHub::checkNick(const string& aNick) {
	string tmp = aNick;
	for(size_t i = 0; i < aNick.size(); ++i) {
		switch(tmp[i]) {
			case '|':
			case '$':
			case '<':
			case '>':
				tmp[i] = '_';
			default:
				continue;
		}
	}
	return tmp;
}

void NmdcHub::connectToMe(const OnlineUser& aUser) {
	checkstate(); char buf[256];
	dcdebug("NmdcHub::connectToMe %s\n", aUser.getIdentity().getNick().c_str());
	ConnectionManager::getInstance()->nmdcExpect(aUser.getIdentity().getNick(), getCurrentNick(), getHubUrl());
	int chars = sprintf_s(buf, 256, "$ConnectToMe %s %s:%d|", fromUtf8(aUser.getIdentity().getNick()).c_str(), getLocalIp().c_str(), ConnectionManager::getInstance()->getPort());
	ConnectionManager::iConnToMeCount++;
	if(chars != -1) send(buf, chars); 
}

void NmdcHub::revConnectToMe(const OnlineUser& aUser) {
	checkstate(); char buf[256]; 
	dcdebug("NmdcHub::revConnectToMe %s\n", aUser.getIdentity().getNick().c_str());
	int chars = sprintf_s(buf, 256, "$RevConnectToMe %s %s|", fromUtf8(getCurrentNick()).c_str(), fromUtf8(aUser.getIdentity().getNick()).c_str());
	Identity bUser = aUser.getIdentity();
	bUser.setFlag(Identity::RCTM_SND);
	if(chars != -1) send(buf, chars);
}

void NmdcHub::hubMessage(const string& aMessage) { 
	checkstate(); 
	char buf[256];
	int iLen = sprintf_s(buf, 256, "<%s> ", getCurrentNick().c_str());
	if(iLen != -1) {
	   send(fromUtf8(string(buf, iLen)+escape(aMessage, true)+"|"));
    }
}
	
void NmdcHub::myInfo() {
	if(state != STATE_LOGGEDIN && state != STATE_MYINFO) {
		return;
	}
	
	reloadSettings(false);
	
	dcdebug("MyInfo %s...\n", getCurrentNick().c_str());
	char StatusMode = '\x01';
	string tmp1 = ";**\x1fU9";
	string tmp2 = "+L9";
	string tmp3 = "+G9";
	string tmp4 = "+R9";
	string tmp5 = "+K9";
	string::size_type i;
	
	for(i = 0; i < 6; i++) {
		tmp1[i]++;
	}
	for(i = 0; i < 3; i++) {
		tmp2[i]++; tmp3[i]++; tmp4[i]++; tmp5[i]++;
	}
	char modeChar = '?';
	switch(SETTING(CONNECTION_TYPE)) {
        case SettingsManager::CONNECTION_ACTIVE:
    		modeChar = 'A';
    		break;
    	case SettingsManager::CONNECTION_PASSIVE:
		    modeChar = 'P';
		    break;
    	case SettingsManager::CONNECTION_SOCKS5:
		    modeChar = '5';
		    break;
		default:
            // ?!?
            break;
    }
            
	
	char tag[256];
	string connection = (BOOLSETTING(USE_UPLOAD_SPEED) || getStealth()) ? SETTING(UPLOAD_SPEED) : SETTING(CONNECTION);
	if (SETTING(THROTTLE_ENABLE) && SETTING(MAX_UPLOAD_SPEED_LIMIT) != 0) {
		sprintf(tag, "%s%s%s%c%s%s%s%s%s%s>", tmp1.c_str(), DCPPVERSIONSTRING, tmp2.c_str(), modeChar, tmp3.c_str(),
		getCounts().c_str(), tmp4.c_str(), Util::toString(Util::getSlots()).c_str(), tmp5.c_str(), Util::toString(SETTING(MAX_UPLOAD_SPEED_LIMIT)).c_str());
	} else {
		sprintf_s(tag, 256, "%s%s%s%c%s%s%s%s>", tmp1.c_str(), DCPPVERSIONSTRING, tmp2.c_str(), modeChar, tmp3.c_str(),
		getCounts().c_str(), tmp4.c_str(), Util::toString(Util::getSlots()).c_str());
	}
	if(getStealth() == false) {
		if(UploadManager::getFireballStatus()) {
            StatusMode |= 0x8;
		} else if(UploadManager::getFileServerStatus()) {
			StatusMode |= 0x4;
		}

		if(Util::getAway()) {
			StatusMode |= 0x2;
		}
	}

	char myinfo[512];
	sprintf_s(myinfo, 512, "$MyINFO $ALL %s %s%s$ $%s%c$%s$", fromUtf8(getCurrentNick()).c_str(), 
		fromUtf8(escape(getCurrentDescription())).c_str(), tag, connection.c_str(), StatusMode, 
		fromUtf8(escape(SETTING(EMAIL))).c_str());
	int64_t newbytesshared = ShareManager::getInstance()->getShareSize();
	if(strcmp(myinfo, lastmyinfo.c_str()) != 0 || newbytesshared < (lastbytesshared - 1048576) || newbytesshared > (lastbytesshared + 1048576)) {
		lastmyinfo = myinfo;
		lastbytesshared = newbytesshared;
		tag[0] = NULL;
		sprintf_s(tag, 256, "%I64d$|", newbytesshared);
		strcat_s(myinfo, 512, tag);
		send(myinfo);
	}
}

void NmdcHub::GetUserList() { 
	OnlineUser::List v;
	
	Lock l(cs);
	for(NickMap::const_iterator i = users.begin(); i != users.end(); ++i) {
		v.push_back(i->second);
	}
	fire(ClientListener::UsersUpdated(), this, v);
}

void NmdcHub::search(int aSizeType, int64_t aSize, int aFileType, const string& aString) {
	checkstate(); 
	AutoArray<char> buf((char*)NULL);
	char c1 = (aSizeType == SearchManager::SIZE_DONTCARE) ? 'F' : 'T';
	char c2 = (aSizeType == SearchManager::SIZE_ATLEAST) ? 'F' : 'T';
	string tmp = ((aFileType == SearchManager::TYPE_TTH) ? "TTH:" + aString : fromUtf8(escape(aString)));
	string::size_type i;
	while((i = tmp.find(' ')) != string::npos) {
		tmp[i] = '$';
	}
	int chars = 0;
	if(ClientManager::getInstance()->isActive() && (!BOOLSETTING(SEARCH_PASSIVE))) {
		string x = ClientManager::getInstance()->getCachedIp();
		const size_t BUF_SIZE = x.length() + aString.length() + 64;
		buf = new char[BUF_SIZE];
		chars = sprintf_s(buf, BUF_SIZE, "$Search %s:%d %c?%c?%I64d?%d?%s|", x.c_str(), (int)SearchManager::getInstance()->getPort(), c1, c2, aSize, aFileType+1, tmp.c_str());
	} else {
		const size_t BUF_SIZE = getCurrentNick().length() + aString.length() + 64;
		buf = new char[BUF_SIZE];
		chars = sprintf_s(buf, BUF_SIZE, "$Search Hub:%s %c?%c?%I64d?%d?%s|", fromUtf8(getCurrentNick()).c_str(), c1, c2, aSize, aFileType+1, tmp.c_str());
	}
	send(buf, chars);
}

string NmdcHub::escape(string str, bool bChat/* = false*/) {
    string::size_type i = 0;

	while( (i = str.find("&amp;", i)) != string::npos) {
		str.replace(i, 1, "&amp;");
		i += 4;
	}
	i = 0;
	while( (i = str.find("&#36;", i)) != string::npos) {
		str.replace(i, 1, "&amp;");
		i += 4;
	}
	i = 0;
	while( (i = str.find("&#124;", i)) != string::npos) {
		str.replace(i, 1, "&amp;");
		i += 4;
	}
	
	if(bChat == false) {
    	i = 0;
    	while( (i = str.find('$', i)) != string::npos) {
    		str.replace(i, 1, "&#36;");
    		i += 4;
    	}
    }
	
	i = 0;
	while( (i = str.find('|', i)) != string::npos) {
		str.replace(i, 1, "&#124;");
		i += 5;
	}
	
	return str;
}

string NmdcHub::unescape(string str) {
    string::size_type i = 0;

    while( (i = str.find("&#36;", i)) != string::npos) {
        str.replace(i, 5, "$");
		i++;
	}
	i = 0;
	while( (i = str.find("&#124;", i)) != string::npos) {
		str.replace(i, 6, "|");
		i++;
	}
	i = 0;
	while( (i = str.find("&amp;", i)) != string::npos) {
		str.replace(i, 5, "&");
		i++;
	}
	
	return str;
}

void NmdcHub::privateMessage(const OnlineUser& aUser, const string& aMessage) { 
	checkstate();

	send("$To: " + fromUtf8(aUser.getIdentity().getNick()) + " From: " + fromUtf8(getCurrentNick()) + " $" + fromUtf8(escape("<" + getCurrentNick() + "> " + aMessage)) + "|");
}

void NmdcHub::clearFlooders(uint64_t aTick) {
	while(!seekers.empty() && seekers.front().second + (5 * 1000) < aTick) {
		seekers.pop_front();
	}

	while(!flooders.empty() && flooders.front().second + (120 * 1000) < aTick) {
		flooders.pop_front();
	}
}

void NmdcHub::on(Connected) throw() {
	Client::on(Connected());

	unsetFlag(SUPPORTS_USERCOMMAND);
	unsetFlag(SUPPORTS_NOGETINFO);
	unsetFlag(SUPPORTS_USERIP2);
	unsetFlag(SUPPORTS_QUICKLIST);

	lastmyinfo.clear();
	lastbytesshared = 0;
	validatenicksent = false;
	bFirstOpList = true;
	PtokaX = false;
    ShittYnHub = false;
}

void NmdcHub::on(Failed, const string& aLine) throw() {
	clearUsers();
	Client::on(Failed(), aLine);
	updateCounts(true);
}

void NmdcHub::on(Second, uint64_t aTick) throw() {
	Client::on(Second(), aTick);
	
	if(state == STATE_LOGGEDIN && (getLastActivity() + getReconnDelay() * 1000) < aTick) {
		send("|", 1);
	} else if(state == STATE_DISCONNECTED && getAutoReconnect() && (aTick > (getLastActivity() + getReconnDelay() * 1000))) {
		// Try to reconnect...
		connect();
	} else if(state == STATE_LOCK && (getLastActivity() + 30000) < aTick) {
        socket->disconnect(false);
    }
}

OnlineUser* NmdcHub::findUserInChat(char * sText, bool &bNormal) {
    string nick = Util::emptyString;
	char *sTmp;
    
    switch(sText[0]) {
        case '<':
            if((sTmp = strchr(sText+1, '>')) != NULL) {
                nick = string(sText+1, sTmp-sText-1);
                if(sTmp[1] == ' ' && sTmp[2] == '/' && tolower(sTmp[3]) == 'm' && tolower(sTmp[4]) == 'e' && sTmp[5] == ' ' && sTmp[6] != NULL) {
                    for(int iNickLen = (sTmp - sText); iNickLen > 1; iNickLen--) {
                        sText[iNickLen+4] = sText[iNickLen-1];
                    }
    				sText[5] = ' ';
    				sText[4] = '*';
    				bNormal = false;
                }
            }
            break;
        case '*':
            if(sText[1] == ' ' && (sTmp = strchr(sText+2, ' ')) != NULL) {
                nick = string(sText+2, sTmp-sText-2);
            }
            break;
        default:
            break;
    }
      
    if(nick != Util::emptyString) {
        return findUser(nick);
    } else {
        return NULL;
    }
}

string NmdcHub::scanNickPrefix(const string& prefix) {
	string maxPrefix;
	tabCompleteNicks.clear();
	{
    	Lock l(cs);
    	for(NickMap::const_iterator i = users.begin(); i != users.end(); ++i) {
    		string prevNick, nick = i->second->getIdentity().getNick(), wholeNick = nick;
    
    		do {
    			string::size_type lp = prefix.size(), ln = nick.size();
    			if((ln >= lp) && (!Util::strnicmp(nick, prefix, lp))) {
    				if(maxPrefix == Util::emptyString)
                        maxPrefix = nick;	//ugly hack
    
    				tabCompleteNicks.push_back(nick);
    				tabCompleteNicks.push_back(wholeNick);
    				maxPrefix = maxPrefix.substr(0, 
                        mismatch(maxPrefix.begin(), maxPrefix.begin()+min(maxPrefix.size(), nick.size()), nick.begin(), compareCharsNoCase).first - maxPrefix.begin());
    			}
    
    			prevNick = nick;
    			nick = stripNick(nick);
    		} while (prevNick != nick);
    	}
    }

	return maxPrefix;
}

string NmdcHub::stripNick(const string& nick) const {
	if(nick.substr(0, 1) != "[")
        return nick;
        
	string::size_type x = nick.find(']');
	string ret;
	// Avoid full deleting of [IMCOOL][CUSIHAVENOTHINGELSETHANBRACKETS]-type nicks
	if((x != string::npos) && (nick.substr(x+1).length() > 0)) {
		ret = nick.substr(x+1);
	} else {
		ret = nick;
	}
	return ret;
}

/* Some parts Copyright (C) 2002-2007 Petr Kozelka, PPK at czdc dot org */