/* 
 * Copyright (C) 2001-2007 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef NMDC_HUB_H
#define NMDC_HUB_H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "TimerManager.h"
#include "SettingsManager.h"

#include "User.h"
#include "CriticalSection.h"
#include "Text.h"
#include "Client.h"
#include "ConnectionManager.h"
#include "UploadManager.h"

#define checkstate() if(state != STATE_LOGGEDIN) return

class ClientManager;

class NmdcHub : public Client, private Flags8
{
public:
	using Client::send;
	using Client::connect;

	virtual void connect(const OnlineUser& aUser, const string&);

	virtual void hubMessage(const string& aMessage);
	virtual void privateMessage(const OnlineUser& aUser, const string& aMessage);
	virtual void sendUserCmd(const string& aUserCmd) throw() { send(fromUtf8(aUserCmd)); }
	virtual void search(int aSizeType, int64_t aSize, int aFileType, const string& aString);
	virtual void password(const string& aPass) { send("$MyPass " + fromUtf8(aPass) + "|"); }
	virtual void info() { myInfo(); }
	virtual void GetUserList();
	
	virtual size_t getUserCount() const {  Lock l(cs); return users.size(); }
	virtual int64_t getAvailable() const;

	virtual string escape(string str, bool bChat = false);
	static string unescape(string str);
	
	virtual OnlineUser* findUser(const string& aNick);
	virtual string scanNickPrefix(const string& prefix);
	virtual string stripNick(const string& nick) const;
private:
	friend class ClientManager;
	enum SupportFlags {
		SUPPORTS_USERCOMMAND  = 0x1,
		SUPPORTS_NOGETINFO    = 0x2,
		SUPPORTS_USERIP2      = 0x4,
		SUPPORTS_QUICKLIST    = 0x8,
	};

	mutable CriticalSection cs;

	typedef HASH_MAP_X(string, OnlineUser*, noCaseStringHash, noCaseStringEq, noCaseStringLess) NickMap;
	typedef NickMap::iterator NickIter;

	NickMap users;

	string lastmyinfo;
	bool validatenicksent, bFirstOpList;
	int64_t lastbytesshared;
	bool PtokaX, ShittYnHub;

	typedef list<pair<string, uint64_t> > FloodMap;
	typedef FloodMap::iterator FloodIter;
	FloodMap seekers;
	FloodMap flooders;

	NmdcHub(const string& aHubURL);	
	virtual ~NmdcHub() throw();

	// Dummy
	NmdcHub(const NmdcHub&);
	NmdcHub& operator=(const NmdcHub&);

	void clearUsers();

	OnlineUser& getUser(const string& aNick);
	void putUser(const string& aNick);

	string toUtf8(const string& str) const { return Text::toUtf8(str); }
	string fromUtf8(const string& str) const { return Text::fromUtf8(str); }
	
	void validateNick(const string& aNick) {
		if(validatenicksent == false) {
			send("$ValidateNick " + fromUtf8(aNick) + "|");
			validatenicksent = true;
		}
	}
	void key(const string& aKey) { send("$Key " + aKey + "|"); }
	void version() { send("$Version 1,0091|"); }
	void getNickList() {
		if(state == STATE_LOGGEDIN || state == STATE_MYINFO) {
			send("$GetNickList|");
		}
	}
	void connectToMe(const OnlineUser& aUser);
	void revConnectToMe(const OnlineUser& aUser);
	void myInfo();
	void supports(const StringList& feat);
	void clearFlooders(uint64_t tick);
	
	void updateFromTag(Identity& id);

	virtual string checkNick(const string& aNick);
	
	OnlineUser* findUserInChat(char * sText, bool &bNormal);
	
	static bool compareCharsNoCase(string::value_type a, string::value_type b) {
		return Text::toLower(a) == Text::toLower(b);
	}

	// TimerManagerListener
	virtual void on(Second, uint64_t aTick) throw();

	virtual void on(Connected) throw();
	virtual void on(Line, const string& l) throw();
	virtual void on(Failed, const string&) throw();
};

#endif // NMDC_HUB_H

/* Some parts Copyright (C) 2002-2007 Petr Kozelka, PPK at czdc dot org */
