/* 
 * Copyright (C) 2004 createagaga
 * Copyright (C) 2004-2007 Petr Kozelka, PPK at czdc dot org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include "../client/DCPlusPlus.h"
#include "AGEmoticonSetup.h"
#include "../client/SimpleXML.h"
#include "../client/pointer.h"

#define COLORREF2RGB(Color) (Color & 0xff00) | ((Color >> 16) & 0xff) | ((Color << 16) & 0xff0000)

CAGEmoticon::CAGEmoticon() {
	m_EmoticonText = Util::emptyStringT;
	m_EmoticonBmp = NULL;
	clrTransparent = 0;
}

CAGEmoticon::~CAGEmoticon() {
	if(m_EmoticonBmp != NULL) {
		DeleteObject(m_EmoticonBmp);
		m_EmoticonBmp = NULL;
	}
}

bool CAGEmoticon::Create(tstring& strEmoticonText, tstring& strEmoticonBmpPath) {
	CDC oTestDC;
	if(oTestDC.CreateCompatibleDC(NULL) == NULL)
		return false;

	m_EmoticonBmp = (HBITMAP) ::LoadImage(0, strEmoticonBmpPath.c_str(), IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
	if(m_EmoticonBmp == NULL) {
		TCHAR buf[1024];
		_stprintf_s(buf, 1024, _T("Unable to load '%s'."), strEmoticonBmpPath.c_str());
		MessageBox(NULL, buf, _T("Unable to load file"), MB_ICONSTOP | MB_OK);
		return false;
	}

	m_EmoticonText = strEmoticonText;

	SelectObject(oTestDC, m_EmoticonBmp);
	clrTransparent = COLORREF2RGB(GetPixel(oTestDC, 0, 0));

	return true;
}

const tstring& CAGEmoticon::GetEmoticonText() {
	return m_EmoticonText;
}

HBITMAP CAGEmoticon::GetEmoticonBmp(COLORREF clrBkColor) {
	HDC BufferDC = CreateCompatibleDC(NULL);
	if(BufferDC == NULL) {  
		return NULL;
	}

	HGDIOBJ PreviousBufferObject = SelectObject(BufferDC, m_EmoticonBmp);
			
	HDC DirectDC = CreateCompatibleDC(NULL);
	if(DirectDC == NULL) {
		DeleteDC(BufferDC);
		return NULL;
	}

	BITMAP bm;
	GetObject(m_EmoticonBmp, sizeof(bm), &bm);
				
	BITMAPINFO RGB32BitsBITMAPINFO; 
	ZeroMemory(&RGB32BitsBITMAPINFO, sizeof(BITMAPINFO));
	RGB32BitsBITMAPINFO.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	RGB32BitsBITMAPINFO.bmiHeader.biWidth = bm.bmWidth;
	RGB32BitsBITMAPINFO.bmiHeader.biHeight = bm.bmHeight;
	RGB32BitsBITMAPINFO.bmiHeader.biPlanes = 1;
	RGB32BitsBITMAPINFO.bmiHeader.biBitCount = 32;

	UINT * ptPixels;	

	HBITMAP DirectBitmap = CreateDIBSection(DirectDC, (BITMAPINFO *)&RGB32BitsBITMAPINFO, DIB_RGB_COLORS, 
		(void **)&ptPixels, NULL, 0);
	if(DirectBitmap == NULL) {
		DeleteDC(DirectDC);
		DeleteDC(BufferDC);
		return NULL;
	}

	HGDIOBJ PreviousObject = SelectObject(DirectDC, DirectBitmap);
	BitBlt(DirectDC, 0, 0, bm.bmWidth, bm.bmHeight, BufferDC, 0, 0, SRCCOPY);
	
	clrBkColor = COLORREF2RGB(clrBkColor);

	for(int i = ((bm.bmWidth*bm.bmHeight)-1); i >= 0; i--) {
		if(ptPixels[i] == clrTransparent)
			ptPixels[i] = clrBkColor;
	}

	SelectObject(DirectDC, PreviousObject);
					
	HBITMAP RetBmp = DirectBitmap;

	DeleteDC(DirectDC);	

	SelectObject(BufferDC, PreviousBufferObject);

	DeleteDC(BufferDC);

    return RetBmp;
}

// CAGEmoticonSetup
CAGEmoticonSetup::CAGEmoticonSetup() {
	setUseEmoticons(false);
}

CAGEmoticonSetup::~CAGEmoticonSetup() {
	for_each(EmoticonsList.begin(), EmoticonsList.end(), DeleteFunction());
}

bool CAGEmoticonSetup::Create(const string& emoticonfile) {
	setUseEmoticons(false);

	if(File::existsFile(emoticonfile) == false)
		return true;

	try {
		SimpleXML xml;
		xml.fromXML(File(emoticonfile, File::READ, File::OPEN).read());
		
		if(xml.findChild("Emotions")) {
			xml.stepIn();

			while(xml.findChild("Emotion")) {
				tstring strEmoticonText = Text::toT(xml.getChildAttrib("ReplacedText"));
				tstring strEmoticonBmpPath = Text::toT(xml.getChildAttrib("BitmapPath"));

				if(strEmoticonBmpPath.size() > 0) {
					strEmoticonBmpPath = Text::toT(Util::getFilePath(emoticonfile)) + strEmoticonBmpPath;
				}

				CAGEmoticon* pEmoticon = new CAGEmoticon();
				if(pEmoticon->Create(strEmoticonText, strEmoticonBmpPath) == false) {
					delete pEmoticon;
					continue;
				}

				EmoticonsList.push_back(pEmoticon);
			}
			xml.stepOut();
		}
	} catch(const Exception& e) {
		dcdebug("CAGEmoticonSetup::Create: %s\n", e.getError().c_str());
		return false;
	}

	if(EmoticonsList.size() == 0) {
		return false;
	}
		
	setUseEmoticons(true);

	return true;
}