/* 
 * Copyright (C) 2003-2007 Petr Kozelka, PPK at czdc dot org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include "../client/DCPlusPlus.h"
#include "Resource.h"
#include "UserInfo.h"

UserInfo::UserInfo(const UserTask& u, bool bUpdate/* = true*/) : UserInfoBase(u.user) {
	for(size_t i = 0; i < COLUMN_LAST; i++) {
		columns[i] = NULL;
	}
	if(bUpdate == true)
	   update(u.identity, -1); 
}

UserInfo::~UserInfo() {
	for(size_t i = 0; i < COLUMN_LAST; i++) {
		if(columns[i] != NULL) {
            delete [] columns[i];
			columns[i] = NULL;
		}
	}
}

const TCHAR* UserInfo::getText(int col) const {
	return columns[col];
}

int UserInfo::compareItems(const UserInfo* a, const UserInfo* b, int col) {
	if(a == NULL || b == NULL){
		return 0;
	}
	
	if(col == COLUMN_NICK) {
		bool aOp = a->getIdentity().isOp();
		bool bOp = b->getIdentity().isOp();
		if(aOp && !bOp) {
			return -1;
		} else if(!aOp && bOp) {
			return 1;
		}
		if(BOOLSETTING(SORT_FAVUSERS_FIRST)) {
			bool a_isFav = FavoriteManager::getInstance()->isFavoriteUser(a->getIdentity().getUser());
			bool b_isFav = FavoriteManager::getInstance()->isFavoriteUser(b->getIdentity().getUser());
			if(a_isFav && !b_isFav)
				return -1;

			if(!a_isFav && b_isFav)
				return 1;
		}
	}
	
	if(col == COLUMN_SHARED || col == COLUMN_EXACT_SHARED) {
		return compare(a->identity.getBytesShared(), b->identity.getBytesShared());
	}
	
	if(a->columns[col] != NULL && b->columns[col] != NULL) {
	   return Util::userstricmp(a->columns[col], b->columns[col]);
    } else if(a->columns[col] == NULL) {
        if(b->columns[col] == NULL) {
            return 0;
        } else {
            return -1;
        }
    } else {
        return 1;
    }
}

bool UserInfo::update(const Identity& identity, int sortCol) {
    bool bNeedSort = (getIdentity().isOp() != identity.isOp());

	tstring old;
	if(sortCol != -1 && columns[sortCol] != NULL)
		old = columns[sortCol];
	
	if(columns[COLUMN_NICK] == NULL) {
		columns[COLUMN_NICK] = Util::tsring2tcharp(columns[COLUMN_NICK], Text::toT(identity.getNick()));
	}

	columns[COLUMN_SHARED] = Util::tsring2tcharp(columns[COLUMN_SHARED], Text::toT(Util::formatBytes(identity.getBytesShared())));
	columns[COLUMN_EXACT_SHARED] = Util::tsring2tcharp(columns[COLUMN_EXACT_SHARED], Text::toT(Util::formatExactSize(identity.getBytesShared())));
	columns[COLUMN_DESCRIPTION] = Util::tsring2tcharp(columns[COLUMN_DESCRIPTION], Text::toT(identity.getDescription()));
	columns[COLUMN_TAG] = Util::tsring2tcharp(columns[COLUMN_TAG], Text::toT(identity.getTag()));
	columns[COLUMN_CONNECTION] = Util::tsring2tcharp(columns[COLUMN_CONNECTION], Text::toT(identity.getConnection()));
	columns[COLUMN_EMAIL] = Util::tsring2tcharp(columns[COLUMN_EMAIL], Text::toT(identity.getEmail()));
	columns[COLUMN_CLIENTID] = Util::tsring2tcharp(columns[COLUMN_CLIENTID], Text::toT(identity.getClientID()));
	columns[COLUMN_VERSION] = Util::tsring2tcharp(columns[COLUMN_VERSION], Text::toT(identity.getVersion()));
	columns[COLUMN_MODE] = Util::tsring2tcharp(columns[COLUMN_MODE], Text::toT(identity.getMode()));
	columns[COLUMN_HUBS] = Util::tsring2tcharp(columns[COLUMN_HUBS], Text::toT(identity.getHubs()));
	columns[COLUMN_SLOTS] = Util::tsring2tcharp(columns[COLUMN_SLOTS], Text::toT(identity.getSlots()));
	columns[COLUMN_LIMITER] = Util::tsring2tcharp(columns[COLUMN_LIMITER], Text::toT(identity.getUpload()));
	columns[COLUMN_IP] = Util::tsring2tcharp(columns[COLUMN_IP], Text::toT(identity.getIp()));

	if(sortCol != -1) {
		bNeedSort = bNeedSort || (columns[sortCol] == NULL && old.size() != 0) || (columns[sortCol] != NULL && old.size() == 0) ||
			((columns[sortCol] != NULL && old.size() != 0) && _tcscmp(old.c_str(), columns[sortCol]) != NULL);
	}
	
    setIdentity(identity);
	return bNeedSort;
}
